//************************************************************************
// 	CaptFrame.CPP
//************************************************************************
//
//  	Copyright (c) 1993-97 by
//
//		ABRI development
//		Andreas Vogel
//		Miltnerweg 17/14/19
//		A-1110 Vienna, Austria
//		phone/fax/modem: +43 1 76 23 35
//		handy:		 +43 664 325 79 70
//		Internet:	 100343.2704@compuserve.com
//
//	All rights reserved
//
// Originally written:
// ===================
// Copyright 1996 Microsoft Systems Journal.
// If this program works, it was written by Paul DiLascia.
// If not, I don't know who wrote it.
// This program compiles with Visual C++ 4.1 on Windows 95
// See ShadeCap.cpp
//
//	--------------------------------------------------------
//
//  Purpose:    Shaded title bar like MS-Excel/WinWord
//
//  Project:
//
//  Author:	Vogel Andreas
//
//  History:
//
//	96-12-xx  V1.0	created
//
//	--------------------------------------------------------
//
//  Brief description:
//
//	editor TAB spacing = 8 !
// 	all my classes start with lower case "c"
// 	all my member variables start with lower case "m"

#include <owl\pch.h>
#pragma hdrstop

#include "captframe.h"

//-------------------------------------------------------- defines -------

#define COLOR_WHITE RGB(255,255,255)
#define COLOR_BLACK RGB(0,0,0)
#define NCOLORSHADES 64		// this many shades in gradient

#define HLSMAX 240	// This is what Display Properties uses
#define RGBMAX 255	// max r/g/b value is 255

//-------------------------------------------------------- globals -------
//------------------------------------------------------ externals -------
//--------------------------------------------- external functions -------
//------------------------------------------------ local functions -------
//-------------------------------------------------------- statics -------

//========================================================================
static void PaintRect(TDC& dc, int x, int y, int w, int h, COLORREF color)
//========================================================================
{
  dc.SelectObject(TBrush(color));
  dc.PatBlt(x, y, w, h, PATCOPY);
  dc.RestoreBrush();
}

//========================================================================
static int GetLuminosity(COLORREF color)
//========================================================================
{
  int r = GetRValue(color);
  int g = GetGValue(color);
  int b = GetBValue(color);
  int rgbMax = max( max(r,g), b);
  int rgbMin = min( min(r,g), b);

  return (((rgbMax+rgbMin) * HLSMAX) + RGBMAX ) / (2*RGBMAX);
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

class cCaptionRect : public TRect

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
// Mini-class to get the caption rectangle of a window in window coords.
// This is the area of the title bar inside the window frame,
// including the icon but not the min/max/close buttons.
{
public:
	cCaptionRect(const TWindow& wnd); // use reference to deny NULL ptr
};

//------------------------------------------------------------------------
cCaptionRect::cCaptionRect(const TWindow& wnd)
//------------------------------------------------------------------------
// cCaptionRect: Constructor conputes caption rectangle in window coords.
{
  // Get size of frame around window
  DWORD dwStyle = wnd.GetStyle();

  TSize szFrame = (dwStyle & WS_THICKFRAME) ?
          TSize(GetSystemMetrics(SM_CXSIZEFRAME),
                     GetSystemMetrics(SM_CYSIZEFRAME)) :
          TSize(GetSystemMetrics(SM_CXFIXEDFRAME),
                          GetSystemMetrics(SM_CYFIXEDFRAME));

//  int cxIcon = GetSystemMetrics(SM_CXSIZE); // width of caption icon/button

  // Compute rectangle
  wnd.GetWindowRect(*this);	// window rect in screen coords

  *this -= TPoint(left, top);	// shift origin to (0,0)

  left  += szFrame.cx;				// frame
  right -= szFrame.cx;				// frame
  top   += szFrame.cy;				// top = end of frame
  bottom = top + GetSystemMetrics(SM_CYCAPTION) // height of caption
               - GetSystemMetrics(SM_CYBORDER); // minus gray shadow border
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

class cRect : public TRect

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
{
public:
	cRect(const int a,const int b,const int c,const int d)
        :
        TRect(a,b,c,d) {}

	cRect(const TRect &r)
        :
        TRect(r) {}

  TRect & DeflateRect(const int a, const int b) {
  	return TRect::Inflate(-a, -b);
  }
};

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
//
//	class cCaptFrame
//
//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

DEFINE_RESPONSE_TABLE1(cCaptFrame, TFrameWindow)
  EV_WM_NCPAINT,
  EV_WM_NCACTIVATE,
  EV_WM_SETTEXT,
  EV_WM_SETTINGCHANGE,
  EV_WM_SYSCOLORCHANGE,
END_RESPONSE_TABLE;

//------------------------------------------------------------------------
void 	cCaptFrame::EvNCPaint(HRGN hRgn)
{
//  TCurrentEvent & cev = GetApplication()->GetCurrentEvent();
//  HRGN hRgn = (HRGN) cev.WParam;

  cCaptionRect rc(*this);		// caption rectangle in window coords
  TRect rcWin;				      // window rect
  GetWindowRect(rcWin);			// .. get window rect
  rc += rcWin.TopLeft();		// convert caption rect to screen coords

  // Don't bother painting if the caption doesn't lie within the region.
  //
  if ((uint)hRgn > 1 && !::RectInRegion(hRgn, &rc))
  {
    TFrameWindow::EvNCPaint(hRgn); // just do default thing
    return;						// and quit
  }

  // Exclude caption from update region
  //
  HRGN hRgnCaption = ::CreateRectRgnIndirect(&rc);
  HRGN hRgnNew     = ::CreateRectRgnIndirect(&rc);
  if((uint)hRgn > 1){
    // wParam is a valid region: subtract caption from it
    ::CombineRgn(hRgnNew, hRgn, hRgnCaption, RGN_DIFF);
  }
  else{
    // wParam is not a valid region: create one that's the whole
    // window minus the caption bar
    HRGN hRgnAll = ::CreateRectRgnIndirect(&rcWin);
    CombineRgn(hRgnNew, hRgnAll, hRgnCaption, RGN_DIFF);
    DeleteObject(hRgnAll);
  }

  // Call Windows to do WM_NCPAINT with altered update region
  //
  TFrameWindow::EvNCPaint(hRgn);	// call it
  DeleteObject(hRgnCaption);	   // clean up
  DeleteObject(hRgnNew);	      // ...

  PaintCaption();	     	// Now paint my special caption
}

//------------------------------------------------------------------------
bool 	cCaptFrame::EvNCActivate(bool  bActive)
//------------------------------------------------------------------------
// Non-client area (de)activated: paint it
{
  // Mimic MFC kludge to stay active if WF_STAYACTIVE bit is on
  //
/*
  #### VOG: Whatever this was good for ...

  if (m_nFlags & WF_STAYACTIVE)
          bActive = TRUE;
*/
  if (!IsWindowEnabled())			// but not if disabled
    bActive = FALSE;

  if (bActive==mbActive)
    return TRUE;					// nothing to do

  // Turn WS_VISIBLE off before callinf DefWindowProc,
  // so DefWindowProc won't paint and thereby cause flicker.
  //
  DWORD dwStyle = GetStyle();

  if (dwStyle & WS_VISIBLE)
    SetWindowLong(GWL_STYLE, (dwStyle & ~ WS_VISIBLE));

  // do not call the base class because it will call Default()
  // and we may have changed bActive.
  DefWindowProc(WM_NCACTIVATE, bActive, 0L);

  if (dwStyle & WS_VISIBLE)
    SetWindowLong(GWL_STYLE, dwStyle);

  // At this point, nothing has happened (since WS_VISIBLE was off).
  // Now it's time to paint.
  //
  mbActive = bActive;			// update state
  SendMessage(WM_NCPAINT);	// paint non-client area (frame too)
  return TRUE;					// done OK
}

//------------------------------------------------------------------------
void	cCaptFrame::EvSetText(CONST char far* text)
//------------------------------------------------------------------------
{
  // Turn WS_VISIBLE style off before calling Windows to
  // set the text, then turn it back on again after.
  //
  DWORD dwStyle = GetStyle();

  if (dwStyle & WS_VISIBLE)
    SetWindowLong(GWL_STYLE, dwStyle & ~ WS_VISIBLE);

  LRESULT lRet = DefWindowProc(WM_SETTEXT, 0, (LPARAM) text);

  if (dwStyle & WS_VISIBLE)
    SetWindowLong(GWL_STYLE, dwStyle);

  mRcCaption.SetNull(); 	// invalidate caption

  SendMessage(WM_NCPAINT);	// paint non-client area (frame too)
  PaintCaption();	        // ..and paint it
}

//------------------------------------------------------------------------
void 	cCaptFrame::EvSysColorChange()
//------------------------------------------------------------------------
{
  delete mFontCaption;
  delete mFontAcme;

  mFontCaption = mFontAcme = NULL;

  mRcCaption.SetNull(); 	// invalidate caption
}

//------------------------------------------------------------------------
void 	cCaptFrame::PaintCaption(bool bActive)
//------------------------------------------------------------------------
{
  if (!IsIconic())
  {
    TWindowDC 	dcWin(*this);
    TMemoryDC 	dc(dcWin);

    dc.SelectObject(*GetCaptionBitmap(bActive));

    const TRect& rc = mRcCaption;

    dcWin.BitBlt(rc.left, rc.top, rc.Width(), rc.Height(), dc, 0, 0, SRCCOPY);
  }
}

//------------------------------------------------------------------------
TBitmap* cCaptFrame::GetCaptionBitmap(bool bActive)
//------------------------------------------------------------------------
{
  TBitmap * & bm = mBmCaption[bActive!=0]; // one of two bitmaps

  cCaptionRect rcCap(*this);  		// caption rectangle in win coords

  if (rcCap != mRcCaption) { 		// if changed since last time:
    delete mBmCaption[0];              // delete both bitmaps,
    delete mBmCaption[1];              // ..they are bad
    mBmCaption[0] = NULL;
    mBmCaption[1] = NULL;
    mRcCaption = rcCap; 	  	// note for next time
  }

  if (bm)
    return bm;		// already have bitmap; return it.

  // Either the size changed or the bitmap was never created. Create it.
  //
  rcCap -= rcCap.TopLeft(); // convert caption rectangle origin to (0,0)
  int w = rcCap.Width();
  int h = rcCap.Height();
  int cxIcon = GetSystemMetrics(SM_CXSIZE);
  int cyIcon = GetSystemMetrics(SM_CYSIZE);

  // Create bitmap same size as caption area and select into memory DC
  //
  TWindowDC 	dcWin(*this);
  TMemoryDC 	dc(dcWin);

  bm = new TBitmap(dcWin, w, h);

  dc.SelectObject(*bm);

  // Paint shaded background. Note all painting is into memory DC,
  // not window, so there's no flicker.
  //
  if (!bActive)
  {
    // Inactive caption: don't do shading, just fill w/bg color
    PaintRect(dc, 0, 0, w, h, GetSysColor(COLOR_INACTIVECAPTION));
  }
  else
  {
    // Active caption: do shading
    //
    COLORREF clrBG = GetSysColor(COLOR_ACTIVECAPTION); // background color
    int r = GetRValue(clrBG);				// red..
    int g = GetGValue(clrBG);				// ..green
    int b = GetBValue(clrBG);				// ..blue color vals
    int x = 5*rcCap.right/6;				// start 5/6 of the way right
    int w = x - rcCap.left;					// width of area to shade
    int xDelta= max(w/NCOLORSHADES,1);	// width of one shade band

    // Paint far right 1/6 of caption the background color
    PaintRect(dc, x, 0, rcCap.right-x, h, clrBG);

    // Compute new color brush for each band from x to x + xDelta.
    // Excel uses a linear algorithm from black to normal, i.e.
    //
    //		color = CaptionColor * r
    //
    // where r is the ratio x/w, which ranges from 0 (x=0, left)
    // to 1 (x=w, right). This results in a mostly black title bar,
    // since we humans don't distinguish dark colors as well as light
    // ones. So instead, I use the formula
    //
    //		color = CaptionColor * [1-(1-r)^2]
    //
    // which still equals black when r=0 and CaptionColor when r=1,
    // but spends more time near CaptionColor. For example, when r=.5,
    // the multiplier is [1-(1-.5)^2] = .75, closer to 1 than .5.
    // I leave the algebra to the reader to verify that the above formula
    // is equivalent to
    //
    //		color = CaptionColor - (CaptionColor*(w-x)*(w-x))/(w*w)
    //
    // The computation looks horrendous, but it's only done once each
    // time the caption changes size; thereafter BitBlt'ed to the screen.
    //
    while (x > xDelta) {			// paint bands right to left
      x -= xDelta;				// next band
      int wmx2 = (w-x)*(w-x);			// w minus x squared
      int w2  = w*w;				// w squared
      PaintRect(dc, x, 0, xDelta, h,
              RGB(r-(r*wmx2)/w2, g-(g*wmx2)/w2, b-(b*wmx2)/w2));
    }

    PaintRect(dc,0,0,x,h,COLOR_BLACK);  // whatever's left ==> black
  }

  // Draw icon and caption buttons.
  // These are all drawn inside a rectangle of dimensions SM_C[XY]SIZE.
  //
  cRect rcButn(0,0,cxIcon,cyIcon);

  // Within the basic button rectangle, Windows 95 uses a 1 or 2 pixel border
  // Icon has 2 pixel border on left, 1 pixel on top/bottom, 0 right
  rcButn.DeflateRect(0,1);
  rcButn.left += 2;

  DrawIconEx(dc.GetHDC(), rcButn.left, rcButn.top,
          GetIconSm(),
          rcButn.Width(), rcButn.Height(), 0, NULL, DI_NORMAL);

  // Close box has a 2 pixel border on all sides but left, which is zero
  rcButn.DeflateRect(0,1);	      		// shrink top/bottom 1 more
  rcButn += TPoint(w-cxIcon-2,0);		// move right

  DrawFrameControl(dc.GetHDC(), &rcButn, DFC_CAPTION, DFCS_CAPTIONCLOSE);

  DWORD dwStyle = GetStyle();

  // Max/restore button is like close box; just shift rectangle left
  //
  if (dwStyle & WS_MAXIMIZEBOX)
  {
    rcButn -= TPoint(cxIcon,0);
    DrawFrameControl(dc.GetHDC(), &rcButn, DFC_CAPTION,
            IsZoomed() ? DFCS_CAPTIONRESTORE : DFCS_CAPTIONMAX);
  }

  // Minimize button has 2 pixel border on all sides but right.
  //
  if (dwStyle & WS_MINIMIZEBOX)
  {
    rcButn -= TPoint(cxIcon-2, 0);
    DrawFrameControl(dc.GetHDC(), &rcButn, DFC_CAPTION, DFCS_CAPTIONMIN);
  }

  // Now draw text. First Create fonts if needed
  //
  if (!mFontCaption)
    CreateFonts();

  // Paint "ACME TEXT" using ACME font, always white
  //
  string s(Title);

  s += " ";

//  CString s = m_strTitle + " ";			 // app title

  cRect rcText = rcCap;				 // caption rectangle

  rcText.left += cxIcon+2;			 // start after icon
  rcText.right = rcButn.left-2;			 // don't draw past buttons
  dc.SetBkMode(TRANSPARENT);			 // draw on top of our shading
  dc.SetTextColor(COLOR_WHITE);			 // always white

  dc.SelectObject(*mFontAcme);

  dc.DrawText(s.c_str(), strlen(s.c_str()), rcText, DT_LEFT|DT_VCENTER|DT_SINGLELINE|DT_END_ELLIPSIS);

  // Now paint window title (caption)
  //
  rcText.left += dc.GetTextExtent(s.c_str(), strlen(s.c_str())).cx; // move past "ACME EDIT "
  if (rcText.right > rcText.left)
  {					// if still room:
    COLORREF clrText;			// text color

    if (bActive)
    {
      // Excel always uses white for title color, but I use the user's
      // selected color--unless it's too dark, then I use white.
      //
      clrText = GetSysColor(COLOR_CAPTIONTEXT);

      if (GetLuminosity(clrText) < 90) // good from trial & error
        clrText = COLOR_WHITE;
    }
    else
        clrText = GetSysColor(COLOR_INACTIVECAPTIONTEXT);

    // Paint the text. Use DT_END_ELLIPSIS to draw ellipsis if text
    // won't fit. Win32 sure is friendly!
    //
    dc.SetTextColor(clrText);
    dc.SelectObject(*mFontCaption);

    //#### document title:
    const char *txt = "ABRI development";
    dc.DrawText(txt, strlen(txt), rcText,
            DT_LEFT|DT_VCENTER|DT_SINGLELINE|DT_END_ELLIPSIS);
  }

  // Restore DC
  dc.RestoreFont();
  dc.RestoreBitmap();

  return bm; // return bitmap to caller--whew!
}

//------------------------------------------------------------------------
void 	cCaptFrame::CreateFonts()
//------------------------------------------------------------------------
// Helper function to build the fonts I need.
{
  // Get current system caption font, just to get its size
  //
  NONCLIENTMETRICS ncm;

  ncm.cbSize = sizeof(ncm);

  SystemParametersInfo(SPI_GETNONCLIENTMETRICS, 0, &ncm, 0);

  mFontCaption = new TFont(&ncm.lfCaptionFont);

  // Create "ACME" font same size as caption font, but use Book Antiqua
  //
  mFontAcme = new TFont("Times New Roman",
  			 ncm.lfCaptionFont.lfHeight,
                         0,0,0,
  			 FW_BOLD);
}

